import numpy as np
import scipy

#1st order terms
class Pk0:    
    def l0(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        expr = D1**2*Pk*(5*b1*(f + 3*xb1) + f*(3*f + 5*xb1))/15
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = D1**2*Pk*(-2*f*sigma*(f*(sigma**2 + 3) + sigma**2*(b1 + xb1))*np.exp(-sigma**2/2) + np.sqrt(2)*np.sqrt(np.pi)*(b1*sigma**4*xb1 + 3*f**2 + f*sigma**2*(b1 + xb1))*Erf(np.sqrt(2)*sigma/2))/(2*sigma**5)
        
        return expr
    
    def l2(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        expr = 2*D1**2*Pk*f*(7*b1 + 6*f + 7*xb1)/21
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = -5*D1**2*Pk*(2*sigma*(3*b1*sigma**4*xb1 + f**2*(2*sigma**4 + 12*sigma**2 + 45) + f*sigma**2*(b1 + xb1)*(2*sigma**2 + 9)) + np.sqrt(2)*np.sqrt(np.pi)*(b1*sigma**4*xb1*(sigma**2 - 3) + 3*f**2*(sigma**2 - 15) + f*sigma**2*(b1 + xb1)*(sigma**2 - 9))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(4*sigma**7)
        
        return expr
    
    
#1st order terms
class GR1:        
    def l1(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = 1j*D1**2*Pk*(-5*b1*xgr1 + 3*f*(gr1 - xgr1) + 5*gr1*xb1)/(5*k1)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = 3*1j*D1**2*Pk*(-2*f*sigma*(gr1 - xgr1)*(sigma**2 + 3) + 2*sigma**3*(b1*xgr1 - gr1*xb1) + np.sqrt(2)*np.sqrt(np.pi)*(3*f*(gr1 - xgr1) + sigma**2*(-b1*xgr1 + gr1*xb1))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(2*k1*sigma**5)
            
        return expr
    
    def l3(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = 2*1j*D1**2*Pk*f*(gr1 - xgr1)/(5*k1)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = 3*1j*D1**2*Pk*(-2*f*sigma*(gr1 - xgr1)*(sigma**2 + 3) + 2*sigma**3*(b1*xgr1 - gr1*xb1) + np.sqrt(2)*np.sqrt(np.pi)*(3*f*(gr1 - xgr1) + sigma**2*(-b1*xgr1 + gr1*xb1))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(2*k1*sigma**5)
        
        return expr
    
    
#2nd order terms
class GR2:        
    def l0(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = D1**2*Pk*(3*b1*xgr2 + f*(gr2 + xgr2) - gr1*xgr1 + 3*gr2*xb1)/(3*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = D1**2*Pk*(-2*sigma*(f*(gr2 + xgr2) + gr1*xgr1) + np.sqrt(2)*np.sqrt(np.pi)*(b1*sigma**2*xgr2 + f*(gr2 + xgr2) + gr1*xgr1 + gr2*sigma**2*xb1)*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(2*k1**2*sigma**3)
        
        return expr
    
    def l2(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = 2*D1**2*Pk*(f*(gr2 + xgr2) + gr1*xgr1)/(3*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = -5*D1**2*Pk*(2*sigma*(3*b1*sigma**2*xgr2 + f*(gr2 + xgr2)*(2*sigma**2 + 9) + 2*gr1*sigma**2*xgr1 + 9*gr1*xgr1 + 3*gr2*sigma**2*xb1) + np.sqrt(2)*np.sqrt(np.pi)*(b1*sigma**4*xgr2 - 3*b1*sigma**2*xgr2 + f*(gr2 + xgr2)*(sigma**2 - 9) + gr1*sigma**2*xgr1 - 9*gr1*xgr1 + gr2*sigma**2*xb1*(sigma**2 - 3))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(4*k1**2*sigma**5)
        
        return expr
   
    
    


import numpy as np
import scipy

class PBBias:
    def __init__(self,cosmo_functions,survey_attr="survey"):
        """
           Gets non-gaussian biases from PBs approach - assumes HMF (Tinker 2010) and HOD (yankelevich and porciani 2018)
        """
        #get which survey out of cosmo_funcs class
        cosmo_func_survey = getattr(cosmo_functions, survey_attr)
        
        def sigma_R_n(R, n,zz=0,cosmo=cosmo, K_MIN = 5e-4,K_MAX=1e+2,steps=int(1e+3)):
            """
            compute sigma^2 for a given radius and n i.e does interal over k
            
            Uses differential equation approach for the yinit
            """
            k = np.logspace(np.log10(K_MIN), np.log10(K_MAX), num=steps)
            pk_L = cosmo_functions.Pk(k)

            def deriv_sigma(x,y,k,Pk,R,n): # from Pylians
                Pkp=np.interp((x),(k),(Pk))
                kR=x*R
                W=3.0*(np.sin(kR)-kR*np.cos(kR))/kR**3 
                return [x**(2+n)*Pkp*W**2]

            #this function computes sigma(R)
            def sigma_sq(k,Pk,R,n):
                k_limits=[k[0],k[-1]]; yinit=[0.0]

                I=scipy.integrate.solve_ivp(deriv_sigma,k_limits,yinit,args=(k,Pk,R,n),method='RK45')

                return I.y[0][-1]#get last value

            sigma_squared = np.zeros((len(R)))
            for i,_ in enumerate(R):
                sigma_squared[i] = sigma_sq(k,pk_L,R[i],n)/(2.0*np.pi**2)

            return sigma_squared
        
        delta_c = 1.686 #from spherical collapse
        
        #so several parameters are 2 dimensional - dependent on R and z
        #therefore we interpolate over redshift here for convenience later where we call things as a fuction of z
        
        #define R range 
        self.R = np.logspace(-3,3,300,dtype=np.float32) #upper limit - causes errors for large R 
        
        # M is the mass enclosed within the radius which also redshift dependent
        self.M_func = lambda xx: ((4*np.pi*cosmo_functions.rho_m(xx)*self.R**3)/3) #so M is in units of solar Mass
        
        #precompute 
        sigmaR0 = sigma_R_n(self.R,0)
        sigmaR1 = sigma_R_n(self.R,-1)
        sigmaR2 = sigma_R_n(self.R,-2)
        
        #for sigma 
        sig_R = {}
        
        sig_R['0'] = lambda xx: sigmaR0* cosmo_functions.D_intp(xx)# (z,R)
        sig_R['1'] = lambda xx: sigmaR1* cosmo_functions.D_intp(xx)
        sig_R['2'] = lambda xx: sigmaR2* cosmo_functions.D_intp(xx)
        
        self.sig_R = sig_R
        
        #peak height (z,R)
        self.nu_func = lambda xx: delta_c/np.sqrt(sig_R['0'](xx))# interpolate along z
        
        #also used in HOD - compute before
        dSdM = CubicSpline(self.M_func(0),np.sqrt(self.sig_R['0'](0))).derivative()(self.M_func(0))
        
        
        #define initial free paremeters from Tinker 2010 table 4, Delta=200
        alpha = 0.368
        beta = 0.589
        gamma = 0.864
        eta = -0.243
        psi = -0.729
        free_args0 = alpha,beta,gamma,eta,psi
        
        def halo_bias_params(zz):#tinker 2010
            alpha,beta0,gamma0,eta0,psi0 = free_args0
            beta = beta0*(1+zz)**(0.2)
            gamma = gamma0*(1+zz)**(-0.01)
            eta = eta0*(1+zz)**(0.27)
            psi = psi0*(1+zz)**(-0.08)
            return alpha,beta,gamma,eta,psi
        
        ######################################################## funcs
        
        #define n_h mean number of halos and N(M,z) mean number of galaxies per halo
        #@njit
        def HOD(zz, M0, NHO):
            #define HoD from Yankalevich and poricani 2018
            M = self.M_func(zz)
            
            theta_HoD = 1 + scipy.special.erf(2*np.log10(M/M0))
            N_c = np.exp(-10*(np.log10(M/M0))**2)+0.05*theta_HoD #central galaxies 
            N_s = 0.003*(M/M0)*theta_HoD

            return NHO*(N_c+N_s)
        
        def multiplicity(zz): # from Tinker 2008,2010
            alpha,beta,gamma,eta,psi = halo_bias_params(zz)
            nu = self.nu_func(zz)
            return alpha*(1+(beta*nu)**(-2*psi))*nu**(2*eta)*np.exp(-gamma*nu**2 /2)
        
        def n_h(zz,free_args=free_args0): #dn/dm
            """
            define halo mass function - number density of halos per unit mass- Tinker2010 
            """
            #derivate of sigma wrt to M
            dSdM = CubicSpline(self.M_func(zz),np.sqrt(self.sig_R['0'](zz))).derivative()(self.M_func(zz))
        
            return (cosmo_functions.rho_m(zz)/self.M_func(zz))*multiplicity(zz)*self.nu_func(zz)*np.abs(dSdM)/np.sqrt(self.sig_R['0'](zz))
        
        self.multiplicity = multiplicity
        self.HMF = n_h
        self.HOD = HOD
        
        #########################################################
        
        # get halo biases these will be arrays with repsect to M - for integration
        def dy_ov_dx(dy,dx):
            """
            Compute the derivative using spline derivatives.
            """
            # Fit cubic splines and caclulate their derivatives for each R coord
            dum1 = CubicSpline(self.R, dy).derivative()(self.R)
            dum2 = CubicSpline(self.R, dx).derivative()(self.R)

            return dum1 / dum2 # Compute the ratio of the derivatives

        #get langrangian and then eulerian biases 
        #see Appendix C ....
        class LagBias:
            """
            define lagrangian biases in terms of z,M and the halo bias params - these are all arrays in (z,M)
            """
            def b1(self,zz):
                alpha,beta,gamma,eta,psi = halo_bias_params(zz)
                nu = self.nu_func(zz)
                return (2*psi)/(delta_c*((beta*nu)**(2*psi)+1)) + (gamma*nu**2-2*eta-1)/delta_c

            def b2(self,zz):
                alpha,beta,gamma,eta,psi = halo_bias_params(zz)
                nu = self.nu_func(zz)
                return (2*psi*(2*gamma*nu**2-4*eta+2*psi-1))/(delta_c**2 *((beta*nu)**(2*psi)+1)) + (gamma**2 *nu**4-4*gamma*eta*nu**2-3*gamma*nu**2+4*eta**2+2*eta)/delta_c**2

        class EulBias:
            """
            define Eulerian biases in terms of lagrangian biases
            """
            def b1(zz,A=1,alpha=0):
                return 1 + LagBias.b1(self,zz)

            def b2(zz,A=1,alpha=0):
                return LagBias.b2(self,zz) - (8/21)*LagBias.b1(self,zz)

            def b_psi(zz,A=1,alpha=0):
                return A*fNL*(2*delta_c*LagBias.b1(self,zz)+4*(dy_ov_dx(np.log(self.sig_R[str(alpha)](zz)),np.log(self.sig_R['0'](zz)))-1))*(self.sig_R[str(alpha)](zz)/self.sig_R['0'](zz))
            
            def b_psi_delta(zz,A=1,alpha=0):
                return A*fNL*(delta_c*(EulBias.b2(zz)+(13/21)*(EulBias.b1(zz)-1))+EulBias.b1(zz)*(2*dy_ov_dx(np.log(self.sig_R[str(alpha)](zz)),np.log(self.sig_R['0'](zz)))-3)+1)*(self.sig_R[str(alpha)](zz)/self.sig_R['0'](zz))
            
        self.EulBias = EulBias
        self.LagBias = LagBias
        
        #########################################################################################
        
        def number_density(zz,M0,NO): 
            """
            return number density for given z 
            """
            
            return np.trapz(HOD(zz,M0,NO)*n_h(zz), (self.M_func(zz)))
        
        # so implement Eq.
        def general_galaxy_bias(b_h,zz,M0,NO,A=1,alpha=0):
            """
            return bias for given z
            """
            
            # Integrate over M for each value of z
            integral_values = np.trapz(b_h(zz, A, alpha)*n_h(zz)*HOD(zz,M0,NO), (self.M_func(zz)))

            return integral_values
        ############################################################################################
        #fit NO and M0 to given n_g and b_1
        def fit_M0(z_arr):
            """
            fit M0 from linear bias (b_1 is independent of NO)
            """
            def objective(M0,zz,NO=1):
                """
                diff between linear bias from PBS and survey specifications
                """
                return general_galaxy_bias(EulBias.b1,zz,M0,NO)/number_density(zz,M0,NO) - cosmo_func_survey.b_1(zz)

            M0_range = (1e+12,1e+14)# should build little tester for this ADD
            #print([objective(M0_range[0],1),objective(M0_range[1],1)])
            M0_arr = np.array([scipy.optimize.brentq(objective, M0_range[0], M0_range[1], args=(z),xtol=1e-4) for i,z in enumerate(z_arr)])

            return CubicSpline(z_arr,M0_arr) # now returns M0 as function of redshift
        
        #now can find NO from n_g
        def fit_NO(z_arr):
            """
            fit NO from number density
            """
            def objective(NO,zz,M0):
                """
                diff between number density from PBS and survey specifications
                """  
                return number_density(zz,M0,NO) - cosmo_func_survey.n_g(zz)

            NO_range = (0,3)# should build little tester for this ADD
            #print([objective(NO_range[0],1,self.M0_func(1)),objective(NO_range[1],1,self.M0_func(1))])
            NO_arr = np.array([scipy.optimize.brentq(objective, NO_range[0], NO_range[1], args=(z,self.M0_func(z)),xtol=1e-4) for i,z in enumerate(z_arr)])

            return CubicSpline(z_arr,NO_arr) # now returns M0 as function of redshift
        
        # so get fittted values of NO and M0
        z_arr = np.linspace(cosmo_functions.z_survey[0],cosmo_functions.z_survey[-1],10)
        self.M0_func = fit_M0(z_arr)
        self.NO_func = fit_NO(z_arr)

        ###############################################################################################
        def get_galaxy_bias(b_h,A=1,alpha=0):
            """
            set z_range and fit cubic spline for a given bias function
            """
            z_samps = np.linspace(cosmo_functions.z_survey[0],cosmo_functions.z_survey[-1],int(40))
            
            bias_arr = np.ones_like(z_samps)
            for i,zz in enumerate(z_samps):
            
                bias_arr[i] = general_galaxy_bias(b_h,zz,self.M0_func(zz),self.NO_func(zz),A,alpha)/cosmo_func_survey.n_g(zz)
            
            return CubicSpline(z_samps,bias_arr)
        
        def get_number_density():
            """
            set z_range and fit cubic spline for a given bias function
            """
            z_samps = np.linspace(cosmo_functions.z_survey[0],cosmo_functions.z_survey[-1],int(40))
            
            n_g = np.ones_like(z_samps)
            for i,zz in enumerate(z_samps):
            
                n_g[i] = number_density(zz,self.M0_func(zz),self.NO_func(zz))
            
            return CubicSpline(z_samps,n_g)
        
        self.get_galaxy_bias = get_galaxy_bias
        
        #################################################################################################
        # so save all required params
        
        self.n_g = get_number_density()
        self.b_1 = get_galaxy_bias(EulBias.b1)
        self.b_2 = get_galaxy_bias(EulBias.b2)
        self.g_2 = lambda xx: -(4/7)*(self.b_1-1)#tidal bias - e.g. baldauf
        
        #get PNG biases for each type
        self.loc = self.Loc(self)
        self.equil = self.Equil(self)
        self.orth = self.Orth(self)
    
    #for the 3 different types of PNG
    class Loc:
        def __init__(self,parent):
            self.A = 1
            self.alpha = 0
            self.b_psi = parent.get_galaxy_bias(parent.EulBias.b_psi,A=self.A,alpha=self.alpha)
            self.b_psi_delta = parent.get_galaxy_bias(parent.EulBias.b_psi_delta,A=self.A,alpha=self.alpha)

    class Equil:
        def __init__(self,parent):
            self.A = 3
            self.alpha = 2
            self.b_psi = parent.get_galaxy_bias(parent.EulBias.b_psi,A=self.A,alpha=self.alpha)
            self.b_psi_delta = parent.get_galaxy_bias(parent.EulBias.b_psi_delta,A=self.A,alpha=self.alpha)

    class Orth:
        def __init__(self,parent):
            self.A = -3
            self.alpha = 1
            self.b_psi = parent.get_galaxy_bias(parent.EulBias.b_psi,A=self.A,alpha=self.alpha)
            self.b_psi_delta = parent.get_galaxy_bias(parent.EulBias.b_psi_delta,A=self.A,alpha=self.alpha)
                



import numpy as np


def get_avg_dist(obs_pos):
    """create x_c to integrate over - lets say it's a 1000 MPc/h (128x128x128) grid situated at (x,y,z)"""
    Nside_theory= 128
    conf_space = np.linspace(0,1000,Nside_theory)
    x_unorm , y_unorm , z_unorm = np.meshgrid(conf_space-obs_pos[0], conf_space-obs_pos[1], conf_space-obs_pos[2],indexing='ij') 
    conf_norm = np.sqrt(x_unorm**2 + y_unorm**2 + z_unorm**2) # make a unit vector - normalise
    ds = np.where(conf_norm==0,1,conf_norm)
    return 1/np.mean(1/ds**3)


def index_tuple(tup,index_cmd):
    """index the array in a tuple"""
    indexed_list = []
    for i,item in enumerate(tup):
        if np.array(item).size>1:
            indexed_list.append(item[index_cmd])
        else:
            indexed_list.append(item)
    return indexed_list  


def tuple_bool(tup,bool_arr):
    """boolean the array in a tuple - when the array are n-dimensional with last two dimensions to 1 """
    indexed_list = []
    for i,item in enumerate(tup):
        if np.array(item).size>1:

            indexed_list.append(item.flatten()[bool_arr][...,None,None])
        else:
            indexed_list.append(item)
    return indexed_list

def get_args(func,params,derivs,betas,r,s,paramsPNG = []):
    """
    get appropiate args for each function type
    """
    func_id = func.__name__ # get string name of func

    args = [params]
    if 'PNG' in func_id:
        args.append(paramsPNG)
    if 'RR' in func_id:
        args.append(derivs)
    if 'GR' in func_id:
        args.append(betas)
    if 'RR' in func_id or 'WA' in func_id:
        args.append(r)
        args.append(s)
    return args


import numpy as np
import scipy


#2nd order terms
class WAGR:
    def l0(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = -2*D1**2*(Pk + Pkd*k1)*(-5*b1*xgr1*(t - 1) + f*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) + 5*gr1*t*xb1)/(15*d*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = D1**2*(2*f*sigma*(2*Pk*(sigma**2 + 6) - 3*Pkd*k1)*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) + 2*sigma**3*(2*Pk - Pkd*k1)*(-b1*xgr1*(t - 1) + gr1*t*xb1) + np.sqrt(2)*np.sqrt(np.pi)*(f*(2*Pk*(sigma**2 - 6) - Pkd*k1*(sigma**2 - 3))*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) - sigma**2*(2*Pk + Pkd*k1*(sigma**2 - 1))*(-b1*xgr1*(t - 1) + gr1*t*xb1))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(2*d*k1**2*sigma**5)
        
        return expr
    
    def l2(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        expr = -2*D1**2*(f*(10*Pk + Pkd*k1)*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) + 7*(2*Pk - Pkd*k1)*(-b1*xgr1*(t - 1) + gr1*t*xb1))/(21*d*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = 5*D1**2*(2*sigma*(f*(2*Pk*(2*sigma**4 + 15*sigma**2 + 90) - 3*Pkd*k1*(sigma**2 + 15))*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) + sigma**2*(2*Pk*(2*sigma**2 + 9) + Pkd*k1*(sigma**2 - 9))*(-b1*xgr1*(t - 1) + gr1*t*xb1)) + np.sqrt(2)*np.sqrt(np.pi)*(-f*(2*Pk*(sigma**4 - 15*sigma**2 + 90) - Pkd*k1*(sigma**4 - 12*sigma**2 + 45))*(gr1*(3*t - 2) - 3*t*xgr1 + xgr1) + sigma**2*(2*Pk*(sigma**2 - 9) + Pkd*k1*(sigma**4 - 4*sigma**2 + 9))*(-b1*xgr1*(t - 1) + gr1*t*xb1))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(4*d*k1**2*sigma**7)
        
        return expr
    
    
class RRGR:
    def l0(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        fd,Dd,_,_,bd1,fdd,Ddd,_,_,bdd1 = cosmo_functions.get_derivs(zz,tracer = cosmo_functions.survey)
        fd,Dd,_,_,xbd1,fdd,Ddd,_,_,xbdd1 = cosmo_functions.get_derivs(zz,tracer = cosmo_functions.survey1)
        
        expr = -D1*(Pk + Pkd*k1)*(D1*(-5*b1*t*xgrd1 + 5*b1*xgrd1 - 5*bd1*t*xgr1 + 3*f*(grd1*t - t*xgrd1 + xgrd1) + 3*fd*gr1*(t - 1) - 3*fd*t*xgr1 + 5*gr1*xbd1*(t - 1) + 5*grd1*t*xb1) + Dd*(2*t - 1)*(-5*b1*xgr1 + 3*f*(gr1 - xgr1) + 5*gr1*xb1))/(15*d*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = D1*(-2*sigma*(D1*(f*(Pk*(sigma**2 + 12) - Pkd*k1*(sigma**2 + 3))*(grd1*t - t*xgrd1 + xgrd1) + fd*(Pk*(sigma**2 + 12) - Pkd*k1*(sigma**2 + 3))*(gr1*(t - 1) - t*xgr1) + sigma**2*(2*Pk - Pkd*k1)*(-b1*t*xgrd1 + b1*xgrd1 - bd1*t*xgr1 + gr1*xbd1*(t - 1) + grd1*t*xb1)) + Dd*(2*t - 1)*(f*(gr1 - xgr1)*(Pk*(sigma**2 + 12) - Pkd*k1*(sigma**2 + 3)) + sigma**2*(2*Pk - Pkd*k1)*(-b1*xgr1 + gr1*xb1))) + np.sqrt(2)*np.sqrt(np.pi)*(D1*(-3*f*(Pk*(sigma**2 - 4) + Pkd*k1)*(grd1*t - t*xgrd1 + xgrd1) - 3*fd*(Pk*(sigma**2 - 4) + Pkd*k1)*(gr1*(t - 1) - t*xgr1) - sigma**2*(Pk*(sigma**2 - 2) + Pkd*k1)*(-b1*t*xgrd1 + b1*xgrd1 - bd1*t*xgr1 + gr1*xbd1*(t - 1) + grd1*t*xb1)) - Dd*(2*t - 1)*(3*f*(gr1 - xgr1)*(Pk*(sigma**2 - 4) + Pkd*k1) + sigma**2*(Pk*(sigma**2 - 2) + Pkd*k1)*(-b1*xgr1 + gr1*xb1)))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(2*d*k1**2*sigma**5)
        
        return expr
    
    def l2(cosmo_functions,k1,zz=0,t=0,sigma=0,nonlin=False):
        k1,Pk,Pkd,Pkdd,d,f,D1 = cosmo_functions.get_params_pk(k1,zz,nonlinear=nonlin)
        
        b1 = cosmo_functions.survey.b_1(zz)
        xb1 = cosmo_functions.survey1.b_1(zz)
        
        gr1,gr2,grd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey)
        xgr1,xgr2,xgrd1,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_ = cosmo_functions.get_beta_funcs(zz,tracer = cosmo_functions.survey1)
        
        fd,Dd,_,_,bd1,fdd,Ddd,_,_,bdd1 = cosmo_functions.get_derivs(zz,tracer = cosmo_functions.survey)
        fd,Dd,_,_,xbd1,fdd,Ddd,_,_,xbdd1 = cosmo_functions.get_derivs(zz,tracer = cosmo_functions.survey1)
        
        expr = 2*D1*(D1*(3*f*(Pk - 2*Pkd*k1)*(grd1*t - t*xgrd1 + xgrd1) + 3*fd*(Pk - 2*Pkd*k1)*(gr1*(t - 1) - t*xgr1) + 7*(2*Pk - Pkd*k1)*(-b1*t*xgrd1 + b1*xgrd1 - bd1*t*xgr1 + gr1*xbd1*(t - 1) + grd1*t*xb1)) + Dd*(2*t - 1)*(3*f*(Pk - 2*Pkd*k1)*(gr1 - xgr1) + 7*(2*Pk - Pkd*k1)*(-b1*xgr1 + gr1*xb1)))/(21*d*k1**2)
        
        Erf = scipy.special.erf
        if sigma != 0:
            expr = 5*D1*(-2*sigma*(D1*(f*(Pk*(2*sigma**4 + 21*sigma**2 + 180) - Pkd*k1*(2*sigma**4 + 12*sigma**2 + 45))*(grd1*t - t*xgrd1 + xgrd1) + fd*(Pk*(2*sigma**4 + 21*sigma**2 + 180) - Pkd*k1*(2*sigma**4 + 12*sigma**2 + 45))*(gr1*(t - 1) - t*xgr1) + sigma**2*(Pk*(sigma**2 + 18) - Pkd*k1*(2*sigma**2 + 9))*(-b1*t*xgrd1 + b1*xgrd1 - bd1*t*xgr1 + gr1*xbd1*(t - 1) + grd1*t*xb1)) + Dd*(2*t - 1)*(f*(gr1 - xgr1)*(Pk*(2*sigma**4 + 21*sigma**2 + 180) - Pkd*k1*(2*sigma**4 + 12*sigma**2 + 45)) + sigma**2*(Pk*(sigma**2 + 18) - Pkd*k1*(2*sigma**2 + 9))*(-b1*xgr1 + gr1*xb1))) + np.sqrt(2)*np.sqrt(np.pi)*(D1*(3*f*(Pk*(sigma**4 - 13*sigma**2 + 60) + Pkd*k1*(sigma**2 - 15))*(grd1*t - t*xgrd1 + xgrd1) + 3*fd*(Pk*(sigma**4 - 13*sigma**2 + 60) + Pkd*k1*(sigma**2 - 15))*(gr1*(t - 1) - t*xgr1) + sigma**2*(Pk*(sigma**4 - 5*sigma**2 + 18) + Pkd*k1*(sigma**2 - 9))*(-b1*t*xgrd1 + b1*xgrd1 - bd1*t*xgr1 + gr1*xbd1*(t - 1) + grd1*t*xb1)) + Dd*(2*t - 1)*(3*f*(gr1 - xgr1)*(Pk*(sigma**4 - 13*sigma**2 + 60) + Pkd*k1*(sigma**2 - 15)) + sigma**2*(Pk*(sigma**4 - 5*sigma**2 + 18) + Pkd*k1*(sigma**2 - 9))*(-b1*xgr1 + gr1*xb1)))*Erf(np.sqrt(2)*sigma/2)*np.exp(sigma**2/2))*np.exp(-sigma**2/2)/(4*d*k1**2*sigma**7)
        
        return expr
 


import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import odeint

# Define the path to the data file relative to the current script location
import os
module_dir = os.path.dirname(os.path.abspath(__file__))
SKAO1Data = np.loadtxt(os.path.join(module_dir, 'data_library/SKAO1Data.txt'))
SKAO2Data = np.loadtxt(os.path.join(module_dir, 'data_library/SKAO2Data.txt'))

class SurveyParams():
    def __init__(self,cosmo):
        """
            gets specifications for some saved surveys (namely linear bias and number density as well as evolution and
            mangification biases
        """

        Euclid = {'b_1':  lambda xx: 0.9 + 0.4*xx,
                    'f_sky': 15000/41253,
                    'z_survey': [0.9,1.8], #get zmin and zmax
                    'be_survey':lambda xx: -7.29 + 0.470*xx + 1.17*xx**2 - 0.290*xx**3,#euclid_data[:,2]
                    'Q_survey': lambda xx: 0.583 + 2.02*xx - 0.568*xx**2 + 0.0411*xx**3,
                    'n_g': lambda zz: 0.0193*zz**(-0.0282) *np.exp(-2.81*zz)}

        #just need D(z)
        baLCDM = cosmo.get_background()
        D_cl = baLCDM['gr.fac. D']
        z_cl = baLCDM['z']
        D_intp = interp1d(z_cl,D_cl,kind='cubic')
        
        BGS =  {'b_1':  lambda xx: 1.34/D_intp(xx),
                'f_sky': 15000/41253,
                'z_survey': [0.05,0.6],
                'be_survey':lambda xx:  -2.25 - 4.02*xx + 0.318*xx**2 - 14.6*xx**3,
                'Q_survey': lambda xx: 0.282 + 2.36*xx + 2.27*xx**2 + 11.1*xx**3,
                'n_g':  lambda zz: 0.023*zz**(-0.471)*np.exp(-5.17*zz)-0.002} #fitting from Maartens


        SKAO1= {'b_1': lambda xx: 0.616*np.exp(1.017*xx),
                'f_sky': 5000/41253,
                'z_survey': [SKAO1Data[:,0][0],SKAO1Data[:,0][-1]],
                'be_survey': interp1d(SKAO1Data[:,0], SKAO1Data[:,4]),
                'Q_survey': interp1d(SKAO1Data[:,0],  SKAO1Data[:,3]),
                'n_g': interp1d(SKAO1Data[:,0], SKAO1Data[:,2])} #fitting from Maartens

        SKAO2= {'b_1': lambda xx: 0.554*np.exp(0.783*xx),
                'f_sky': 30000/41253,
                'z_survey':  [SKAO2Data[:,0][0],SKAO2Data[:,0][-1]],
                'be_survey': interp1d(SKAO2Data[:,0], SKAO2Data[:,4]),
                'Q_survey': interp1d(SKAO2Data[:,0],  SKAO2Data[:,3]),
                'n_g': interp1d(SKAO2Data[:,0], SKAO2Data[:,2])} #fitting from Maartens
        
        DM = {'b_1': lambda xx: 1 + 0*xx ,   # for dark matter particles 
              'b_2': lambda xx:  0*xx ,
              'g_2': lambda xx:  0*xx ,
              'f_sky': 1,
              'z_survey': [0,3],
              'be_survey':lambda xx:  0*xx ,
              'Q_survey': lambda xx: 0*xx,
              'n_g':  lambda zz: 1e+5 + 0*xx}
        
        self.survey_dict= {'Euclid':Euclid,
                             'BGS':BGS,
                             'SKAO1':SKAO1,
                             'SKAO2':SKAO2,
                             'DM':DM}
        
        ################################################################################
        
    def get_surveybias(self,survey):
        
        return self.survey_dict[survey]
    
class SurveyFunctions:
    """ read in survey specific params from dict - if empty use default values """
    def __init__(self,survey_params,all_bias=True):
        # set evolution and magnification bias
        if 'be_survey' in survey_params.keys():
            self.be_survey = survey_params['be_survey']
        else:
            self.be_survey  =  lambda xx: 0*xx

        if 'Q_survey' in survey_params.keys():
            self.Q_survey = survey_params['Q_survey']
        else:
            self.Q_survey =  lambda xx: 0*xx #=2/5

        if all_bias:
            # define other survey specific params not set by HOD and HMF
            if 'n_g' in survey_params.keys():
                self.n_g = survey_params['n_g']
            else:
                self.n_g  =  lambda xx: 1e+5 + 0*xx

            #define default bias stuff
            if 'b_1' in survey_params:
                self.b_1 = survey_params['b_1']
            else:
                self.b_1  =  lambda xx: np.sqrt(1+xx)

            if 'be_survey' in survey_params.keys():
                self.be_survey = survey_params['be_survey']
            else:
                self.be_survey  =  lambda xx: 0*xx

            if 'Q_survey' in survey_params.keys():
                self.Q_survey = survey_params['Q_survey']
            else:
                self.Q_survey =  lambda xx: 0*xx #=2/5

            #define second order biases from b_1 # or can define them in dict as above 

            if 'g_2' in survey_params.keys():
                self.g_2 = survey_params['g_2']
            else:
                self.g_2 = lambda xx: -(4/7)*(self.b_1(xx)-1) # 0.524-0.547*self.b_1(xx)+0.046*self.b_1(xx)**2#

            if 'b_2' in survey_params.keys():
                self.b_2 = survey_params['b_2']
            else:    
                self.b_2 = lambda xx: 0.412 - 2.143*self.b_1(xx) +0.929*self.b_1(xx)**2 + 0.008*self.b_1(xx)**3 + 4/3 * self.g_2(xx) 
            
        ###########################################################
        #for non tracer things  
        if 'f_sky' in survey_params.keys():
            self.f_sky = survey_params['f_sky']
        else:
            self.f_sky  =  1   

        if 'z_survey' in survey_params.keys():
            self.z_survey = np.linspace(survey_params['z_survey'][0],survey_params['z_survey'][1],int(1e+5))  
        else:
            self.z_survey  = np.linspace(0,3,int(1e+5))



#for numerical angular derivates - usfule for FOG and consistency otherwise precompute analytic are quicker
def bk_ylm(func,l,m,params,derivs,betas,r=0,s=0,sigma=None,paramsPNG=[]):
    """
    does numerical integration with double legendre gauss integration implemented in int_gl_dbl()
    is vectorised just last two axes need to be dimension 1 
    """

    def int_gl_dbl(func,*args):
        """
        implements double legendre guass integral for mu and phi integrals - extension and specialiation
        of integrate.fixed_quad()
        """
        n=16

        nodes, weights = np.polynomial.legendre.leggauss(n)#legendre gauss - get nodes and weights for given n
        nodes = np.real(nodes)
        mesh_nodes1,mesh_nodes2 = np.meshgrid(nodes,nodes,indexing ='ij')        #mesh gridding as 2d
        mesh_weights1,mesh_weights2 = np.meshgrid(weights,weights,indexing ='ij')

        phi_nodes = (2*np.pi)*(mesh_nodes1+1)/2.0 #sample phi range [0,2*np.pi]
        mu_nodes = (2)*(mesh_nodes2+1)/2.0 - 1 # sample mu range [-1,1] - so this is the natural gauss legendre range!

        return (2*np.pi)/2.0 * np.sum(mesh_weights1*mesh_weights2*func(phi_nodes, mu_nodes, *args), axis=(-2,-1)) #sum over last two axes (mu and phi)
    
    k1,k2,k3,theta,Pk1,Pk2,Pk3,Pkd1,Pkd2,Pkd3,Pkdd1,Pkdd2,Pkdd3,d,K,C,f,D1,b1,b2,g2 = params
    def integrand(phi,mu,*args):
        ylm = scipy.special.sph_harm(m, l, phi, np.arccos(mu))
        expression = func(*args,mu,phi)
        
        if sigma is None: #no FOG
            dfog_val = 1
        else:
            mu2 = mu*np.cos(theta)+ (1-mu**2)**(1/2) *np.sin(theta)*np.cos(phi)
            mu3 = -(mu*k1+mu2*k2)/k3
            dfog_val = np.exp(-(1/2)*((k1*mu)**2+(k2 *mu2)**2+(k3 *mu3)**2)*sigma**2)
            
        return np.conjugate(ylm)*expression*dfog_val
    
    args = get_args(func,params,derivs,betas,r,s,paramsPNG=paramsPNG) # get necessary args for the func
    
    result = int_gl_dbl(integrand,*args)
        
    return result


import numpy as np
import scipy
from scipy.interpolate import interp1d
from scipy.integrate import odeint

from cosmo_wap.peak_background_bias import *
from cosmo_wap.survey_params import *

#from useful_funcs import *            

class CosmoWAP:
    """
           Main functions - takes in cosmology from CLASS and bias models and then can called to generate cosmology (f,P(k),P'(k),D(z) etc) and all other biases including relativstic parts
    
    """
    def __init__(self,cosmo,survey_params={},compute_bias=True):
        """
           Inputs CLASS and bias dict to return all bias and cosmological parameters defined within the class object

        """
        self.cosmo = cosmo
        baLCDM = cosmo.get_background()
        
        D_cl = baLCDM['gr.fac. D']
        z_cl = baLCDM['z']
        self.D_intp = interp1d(z_cl,D_cl,kind='cubic')
        
        f_cl = baLCDM['gr.fac. f']
        D_cl = baLCDM['gr.fac. D']
        z_cl = baLCDM['z']
        H_cl = baLCDM['H [1/Mpc]']
        xi_cl = baLCDM['comov. dist.']
        t_cl = baLCDM['conf. time [Mpc]']
        
        self.z_cl = z_cl#save for later
        self.Om_0 = cosmo.get_current_derived_parameters(['Omega_m'])['Omega_m']
        self.h = cosmo.get_current_derived_parameters(['h'])['h']
        
        #define functions and add in h
        self.H_c = interp1d(z_cl,H_cl*(1/(1+z_cl))/self.h,kind='cubic') # now in h/Mpc!#is conformal
        self.dH_c = interp1d(z_cl,np.gradient(H_cl*(1/(1+z_cl))/self.h,z_cl),kind='cubic') # derivative wrt z
        self.ddH_c = interp1d(z_cl,np.gradient(self.dH_c(z_cl),z_cl),kind='cubic') # second derivative wrt z
        self.comoving_dist = interp1d(z_cl,xi_cl*self.h,kind='cubic') # just use class background as quick
        self.d_to_z = interp1d(xi_cl*self.h,z_cl,kind='cubic') # useful to map other way
        self.f_intp = interp1d(z_cl,f_cl,kind='cubic')#get f #omega_mz = Omega_m *(1+zt)**3 /(Omega_m *(1+zt)**3 + Omega_l)
        self.D_intp = interp1d(z_cl,D_cl,kind='cubic')
        self.dD_dz = interp1d(z_cl,np.gradient(D_cl,z_cl),kind='cubic')
        self.conf_time = interp1d(z_cl,self.h*t_cl,kind='cubic')#convert between the two
        #misc
        self.c = 2.99792e+5 #km/s
        self.H0 = 100/self.c #[h/Mpc]
        self.Om = lambda xx: self.Om_0 * (self.H0**2 / self.H_c(xx)**2) * (1+xx)
        
        #for critical density
        GG = 4.300917e-3 #[pc SolarMass (km/s)^2]
        self.G = GG/(1e+6*self.c**2)#gravitational constant
        self.rho_crit = lambda xx: 3*self.H_c(xx)**2/(8*np.pi*self.G)  #in units of h^3 Mo/ Mpc^3 where Mo is solar mass
        self.rho_m = lambda xx: self.rho_crit(xx)*self.Om(xx)          #in units of h^3 Mo/ Mpc^3
        
        #######################################################################################################
       
        #setup surveys and compute all bias params including for multi tracer case...        
        self = self.setup_survey(survey_params, compute_bias, cosmo_functions)
               
        # get bias funcs for given surveys
        if len(survey_params) == 1:
            survey_params = survey_params[0]

        if compute_bias:
            print("Computing bias params...")
            self.survey = PBBias(cosmo_functions)
            self.survey = SurveyFunctions(survey_params,False)
        else:
            self.survey = SurveyFunctions(survey_params)

        try survey1 = survey_params[1]:
            if compute_bias:
                print("Computing bias params...")
                self.survey1 = PBBias(cosmo_functions,"survey1")
                self.survey1 = SurveyFunctions(survey1,False)
            else:
                self.survey1 = SurveyFunctions(survey1)
        except:
            self.survey1 = self.survey

        #get ranges of cross survey
        z_min = max([self.survey.z_survey[0],self.survey1.z_survey[0]])
        z_max = min([self.survey.z_survey[1],self.survey1.z_survey[1]])
        if z_min > z_max:
            raise ValueError("incompatible survey redshifts.")
        self.z_survey = np.linspace(z_min,z_max,int(1e+5))
        self.f_sky = min([self.survey.f_sky,self.survey1.f_sky])

        ####################################################################################################
        #get power spectras
        def get_class_powerspectrum(kk,zz=0): #h are needed to convert to 1/Mpc for k then convert pk back to (Mpc/h)^3
            return np.array([self.cosmo.pk_lin(ki, zz) for ki in kk*self.h])*self.h**3
    
        def get_Pk_NL(k,z=0): # for halofit non-linear power spectrum
            pk_nl = interp1d(k,np.array([self.cosmo.pk(ki, z) for ki in k*self.h])*self.h**3)
            return pk_nl

        def get_pkinfo_z(k,z):
            """get Pk and its k derivatives"""
            Plin = get_class_powerspectrum(k,0)#just always get present day power spectrum
            Pk = interp1d(k,Plin)#get linear power spectrum
            Pk_d = interp1d(k,np.gradient(Plin,k))
            Pk_dd = interp1d(k,np.gradient(Pk_d(k),k))
            return Pk,Pk_d,Pk_dd
        
        K_MAX = 100
        k = np.logspace(-5, np.log10(K_MAX), num=1000)
        self.Pk,self.Pk_d,self.Pk_dd = get_pkinfo_z(k,0)
        self.Pk_L = get_class_powerspectrum
        self.Pk_NL = get_Pk_NL(k,0) #if you want HALOFIT P(k)
            
    #getter functions and their requisites    
    ###########################################################
    def _process_survey(self, survey, compute_bias, cosmo_functions, label="survey"):
        if compute_bias:
            print(f"Computing bias params for {label}...")
            survey_bias = PBBias(cosmo_functions, label)
            return SurveyFunctions(survey, False)
        else:
            return SurveyFunctions(survey)

    # Get bias functions for given surveys allowing for two surveys...
    def setup_survey(self, survey_params, compute_bias, cosmo_functions):
        # If survey_params is a list 
        if len(survey_params) > 0:
            survey = survey_params[0]
        else:
            survey = survey_params

        # Process first survey (or only survey)
        self.survey = self._process_survey(survey, compute_bias, cosmo_functions)

        # Check for additional surveys
        if len(survey_params) > 1:
            self.survey1 = self._process_survey(survey_params[1], compute_bias, cosmo_functions, label="survey1")
        else: #is same as first
            self.survey1 = self.survey
            
        return self
                
    #######################################################################################################
    
    def Pk_phi(self,k, k0=0.05, units=True):
        """Power spectrum of the Bardeen potential Phi in the matter-dominated era - k in units of h/Mpc.
        """
        k_pivot = k0/self.h
        resp = (9.0/25.0) * self.cosmo.get_current_derived_parameters(['A_s'])['A_s'] * (k/k_pivot)**(self.cosmo.get_current_derived_parameters(['n_s'])['n_s'] - 1.0)

        if units:
            resp *= 2*np.pi**2.0/k**3.0    #[Mpc/h]^3

        return resp

    def M(self,k, z):
        """The scaling factor between the primordial scalar power spectrum and the late-time matter power spectrum
        """
        return np.sqrt(self.D_intp(z)**2 *self.Pk(k) / self.Pk_phi(k))

    def solve_second_order_KC(self):
        c = 2.99792*10**5 #km/s
        def F_func(u,zz): # so variables are F and H and D
            f,fd = u # unpack u vector
            D_zz = self.D_intp(zz)
            return [fd,(-self.H_c(zz)*self.dH_c(zz)*(1+zz)**2 *fd + ((3*(self.H0)**2 * self.Om_0 *(1+zz))/(2))*(f+D_zz**2))/(self.H_c(zz)**2 *(1+zz)**2)]

        odeint_zz = np.linspace(20,0.05,int(1e+5))# so z=20 should be pretty much matter dominated

        #set initial params for F
        F0 = [(3/7)*self.D_intp(odeint_zz[0])**2,(3/7)*2*self.D_intp(odeint_zz[0])*self.dD_dz(odeint_zz[0])]
        sol1 = odeint(F_func,F0,odeint_zz)
        K = (sol1[:,0]/self.D_intp(odeint_zz)**2)
        C = sol1[:,1]/(2*self.D_intp(odeint_zz)*self.dD_dz(odeint_zz))
        self.K_intp = interp1d(odeint_zz,K,kind='cubic')
        self.C_intp = interp1d(odeint_zz,C,kind='cubic')

    def lnd_derivatives(self,functions_to_differentiate):
        """
            calculates derivatives of a list of functions wrt log comoving dist numerically
        """

        # Store first derivatives in a list
        function_derivatives = []

        for func in functions_to_differentiate:
            # Calculate numerical derivatives of the function with respect to ln(d)
            derivative_func = interp1d(self.z_survey, np.gradient(func(self.z_survey), np.log(self.comoving_dist(self.z_survey))))
            function_derivatives.append(derivative_func)

        return function_derivatives

    def get_params(self,k1,k2,k3=None,theta=None,zz=0,tracer = None,nonlin=False,growth2=False):
        """
            return arrays of redshift and k dependent parameters
        """
        if theta is None:
            if k3 is None:
                raise  ValueError('Define either theta or k3')
            else:
                theta = np.arccos((k3**2 - k1**2 - k2**2)/(2*k1*k2))
                
        if tracer is None:
            tracer = self.survey
        
        k3 = np.sqrt(k1**2 + k2**2 + 2*k1*k2*np.cos(theta))#get k3 from triangle condition
        k3 = np.where(k3==0,1e-4,k3)
        
        Pk1 = self.Pk(k1)
        Pk2 = self.Pk(k2)
        Pk3 = self.Pk(k3)

        Pkd1 = self.Pk_d(k1)
        Pkd2 = self.Pk_d(k2)
        Pkd3 = self.Pk_d(k3)

        Pkdd1= self.Pk_dd(k1)
        Pkdd2 = self.Pk_dd(k2)
        Pkdd3 = self.Pk_dd(k3)
        
        if nonlin:
            Pk1 = Pk_NL(k1)
            Pk2 = Pk_NL(k2)
            Pk3 = Pk_NL(k3)
        
        #redshift dependendent terms
        d = self.comoving_dist(zz)
        
        K = 3/7 # from einstein-de-sitter
        C = 3/7
        if growth2:
            self.solve_second_order_KC()#get K and C
            K = self.K_intp(zz)
            C = self.C_intp(zz)

        f = self.f_intp(zz)
        D1 = self.D_intp(zz)
        
        #survey stuff
        b1 = tracer.b_1(zz)
        b2 = tracer.b_2(zz)
        g2 = tracer.g_2(zz)
        return k1,k2,k3,theta,Pk1,Pk2,Pk3,Pkd1,Pkd2,Pkd3,Pkdd1,Pkdd2,Pkdd3,d,K,C,f,D1,b1,b2,g2
    
    def get_params_pk(self,k1,zz,nonlinear=False):
        """
           return arrays of redshift and k dependent parameters but just for single knitting vector
        """
        
        Pk1 = self.Pk(k1)
        Pkd1 = self.Pk_d(k1)
        Pkdd1= self.Pk_dd(k1)
        
        if nonlinear:
            Pk1 = Pk_NL(k1)
            
        #redshift dependendent terms
        d = self.comoving_dist(zz)


        f = self.f_intp(zz)
        D1 = self.D_intp(zz)
        
        return k1,Pk1,Pkd1,Pkdd1,d,f,D1

    def get_derivs(self,zz,tracer = None):
        """
        Derivatives of redshift dependent parameters for radial wide separation terms
        """
        if tracer is None:
            tracer = self.survey
            
        #get derivs of these redshift dependent functions
        func_deriv_list = self.lnd_derivatives([tracer.b_1,tracer.b_2,tracer.g_2,self.f_intp,self.D_intp])
        tracer.b1_d,tracer.b2_d,tracer.g2_d,self.f_d,self.D_d = func_deriv_list
        func_second_deriv_list = self.lnd_derivatives(func_deriv_list)
        tracer.b1_dd,tracer.b2_dd,tracer.g2_dd,self.f_dd,self.D_dd = func_second_deriv_list
        
        #1st deriv
        fd = self.f_d(zz)
        Dd = self.D_d(zz)
        gd2 = tracer.g2_d(zz)
        bd2 = tracer.b2_d(zz)
        bd1 = tracer.b1_d(zz)
        #2nd deriv
        fdd = self.f_dd(zz)
        Ddd = self.D_dd(zz)
        gdd2 = tracer.g2_dd(zz)
        bdd2 = tracer.b2_dd(zz)
        bdd1 = tracer.b1_dd(zz)
        return fd,Dd,gd2,bd2,bd1,fdd,Ddd,gdd2,bdd2,bdd1
    
    def get_PNGparams(self,fNL,zz,k1,k2,k3,tracer = None):
        """
        returns terms needed for local type fnl including scale dependent bias
        """
        if tracer is None:
            tracer = self.survey
            
        delta_c = 1.686
        bL10 = lambda xx: tracer.b_1(xx) - 1
        bL20 = lambda xx: tracer.b_2(xx) - (8/21)*bL10(xx)

        bL01 = lambda xx: 2*delta_c*fNL*bL10(xx)
        bL11 = lambda xx: 2*fNL*(delta_c*bL20(xx)-bL10(xx))
        bL02 = lambda xx: 4*fNL**2 * delta_c*(delta_c*bL20(xx)-2*bL10(xx))

        bE01 = bL01
        bE11 = lambda xx: bL01(xx) + bL11(xx)
        bE02 = bL02

        alphak1 = self.M(k1, 0)
        alphak2 = self.M(k2, 0)
        alphak3 = self.M(k3, 0)
        return fNL,1*bE01(zz),1*bE11(zz),1*bE02(zz),alphak1,alphak2,alphak3
    
    def get_beta_funcs(self,zz,tracer = None):
        """
            Function that relies on biases and functions defined above to return beta coefficients (beta expressions adapted from Eline de Weerd GitHub) from paper 1711.01812v4
        """
        if tracer is None:
            tracer = self.survey
            
        #these derivs here are wrt to conformal time so we convert to derivs wrt to z
        #d/dt = da/dt d/da = a H dz/da d/dz =  -(1+z) H d/dz # everything here is conformal both t and H
        #d^2/d^2 t = (1+z)^2 H^2 d^2/d z^2 + H(1+z)(H+(1+z)H')d/dz
        
        dQ_dz = interp1d(self.z_survey,np.gradient(tracer.Q_survey(self.z_survey),self.z_survey))
        dbe_dz = interp1d(self.z_survey,np.gradient(tracer.be_survey(self.z_survey),self.z_survey))
        db1_dz = interp1d(self.z_survey,np.gradient(tracer.b_1(self.z_survey),self.z_survey))
        
        dH_dt = lambda xx: -(1+xx)*self.H_c(xx)*self.dH_c(xx)
        dH_dt2 = lambda xx: (1+xx)**2 *self.H_c(xx)**2 *self.ddH_c(xx)+self.H_c(xx)*(1+xx)*(self.H_c(xx)+(1+xx)*self.dH_c(xx))*self.dH_c(xx)#should just check with doing it numerically as is easier
        #self.dH22 = interp1d(self.z_cl,np.gradient(dH_dt(self.z_cl),self.conf_time(self.z_cl)))
        self.dH_dt = dH_dt                   
        dQ_dt = lambda xx: -(1+xx)*self.H_c(xx)*dQ_dz(xx)
        dbe_dt = lambda xx: -(1+xx)*self.H_c(xx)*dbe_dz(xx)#0*xx-1.6*1e-4
        db1_dt = lambda xx: -(1+xx)*self.H_c(xx)*db1_dz(xx)  
        
        self.Q = tracer.Q_survey  # bit silly but whatever
        self.b_e = tracer.be_survey

        partdQ=0
        partdb1=0
        
        #from 1st order petrubation theory
        tracer.gr1 = lambda xx: self.H_c(xx)*self.f_intp(xx)*(self.b_e(xx)-2*self.Q(xx)-2*(1-self.Q(xx))/(self.comoving_dist(xx)*self.H_c(xx))-dH_dt(xx)/self.H_c(xx)**2)
        tracer.gr2 = lambda xx: self.H_c(xx)**2 *(self.f_intp(xx)*(3-self.b_e(xx))+ (3/2)*self.Om(xx)*(2+self.b_e(xx)-self.f_intp(xx)-4*self.Q(xx)-2*self.Q(xx)-2*(1-self.Q(xx))/(self.comoving_dist(xx)*self.H_c(xx))-dH_dt(xx)/self.H_c(xx)**2))
        tracer.grd1 = self.lnd_derivatives([tracer.gr1])[0]#get derivative for 1st order coef
        
        beta = np.empty(20,dtype=object)
        beta[6] = lambda xx: self.H_c(xx)**2 * ((3 / 2) * self.Om(xx) * (2 - 2 * self.f_intp(xx)+ self.b_e(xx) - 4 * self.Q(xx) - ((2 * (1 - self.Q(xx))) / (self.comoving_dist(xx) * self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2)))
        beta[7] = lambda xx: self.H_c(xx)**2 * (self.f_intp(xx)* (3 - self.b_e(xx)))
        beta[8] = lambda xx: self.H_c(xx)**2 * (3 * self.Om(xx) * self.f_intp(xx) * (2 - self.f_intp(xx) - 2 * self.Q(xx)) + self.f_intp(xx)**2 * (4 + self.b_e(xx) - self.b_e(xx)**2 + 4 * self.b_e(xx) * self.Q(xx) - 6 * self.Q(xx) - 4 * self.Q(xx)**2 + 4 * partdQ + 4 * (dQ_dt(xx) / self.H_c(xx)) -(dbe_dt(xx) / self.H_c(xx)) - (2 / (self.comoving_dist(xx)**2 * self.H_c(xx)**2)) * (1 - self.Q(xx) + 2 * self.Q(xx)**2 - 2 * partdQ) - (2 / (self.comoving_dist(xx) * self.H_c(xx))) * (3 - 2 * self.b_e(xx) + 2 * self.b_e(xx) * self.Q(xx) - self.Q(xx) - 4 * self.Q(xx)**2 + ((3 * dH_dt(xx)) / (self.H_c(xx)**2)) * (1 - self.Q(xx)) + 4 * partdQ + 2 * (dQ_dt(xx) / self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2) * (3 - 2 * self.b_e(xx) + 4 * self.Q(xx) + ((3 * dH_dt(xx)) / (self.H_c(xx)**2))) + (dH_dt2(xx) / self.H_c(xx)**3)))
        beta[9] = lambda xx: self.H_c(xx)**2 * ( -(9 / 2) * self.Om(xx) * self.f_intp(xx))
        beta[10] = lambda xx: self.H_c(xx)**2 * (3 * self.Om(xx) * self.f_intp(xx))
        beta[11] = lambda xx: self.H_c(xx)**2 * ( (3/2 ) * self.Om(xx) * (1 + 2 * self.f_intp(xx)/ (3 * self.Om(xx))) + 3 * self.Om(xx) * self.f_intp(xx)- self.f_intp(xx)**2 * (-1 + self.b_e(xx) - 2 * self.Q(xx) - ((2 * (1 + self.Q(xx))) / (self.comoving_dist(xx) * self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2)))
        beta[12] = lambda xx: self.H_c(xx)**2 * ( -3 * self.Om(xx) * (1 + 2 * self.f_intp(xx)/ (3 * self.Om(xx))) - self.f_intp(xx)* ( tracer.b_1(xx) * (self.f_intp(xx)- 3 + self.b_e(xx)) + (db1_dt(xx) / self.H_c(xx)) ) + (3 / 2) * self.Om(xx) * (tracer.b_1(xx) * (2 + self.b_e(xx) - 4 * self.Q(xx) - 2 * ((1 - self.Q(xx))/(self.comoving_dist(xx) * self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2) ) + db1_dt(xx) /self.H_c(xx) + 2 * (2 - (1 / (self.comoving_dist(xx) * self.H_c(xx))) ) * partdb1 ) )    
        beta[13] = lambda xx: self.H_c(xx)**2 * (( (9 / 4) * self.Om(xx)**2 + (3 / 2) * self.Om(xx) * self.f_intp(xx)* (1 - (2 * self.f_intp(xx)) + 2 * self.b_e(xx) - 6 * self.Q(xx) - ((4 * (1 - self.Q(xx)))/(self.comoving_dist(xx) * self.H_c(xx))) - ((3 * dH_dt(xx)) / self.H_c(xx)**2) ) ) + ( self.f_intp(xx)**2 * (3 - self.b_e(xx)) ) )
        beta[14] = lambda xx: self.H_c(xx) * ( - (3 / 2) * self.Om(xx) * tracer.b_1(xx))
        beta[15] = lambda xx: self.H_c(xx) * 2 * self.f_intp(xx)**2
        beta[16] = lambda xx: self.H_c(xx) * (self.f_intp(xx)* (tracer.b_1(xx) * (self.f_intp(xx)+ self.b_e(xx) - 2 * self.Q(xx) - ((2 * (1 - self.Q(xx))) / (self.comoving_dist(xx) * self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2)) + (db1_dt(xx) / self.H_c(xx)) + 2 * (1 - (1 / (self.comoving_dist(xx) * self.H_c(xx)))) * partdb1 ))
        beta[17] = lambda xx: self.H_c(xx) * (- (3 / 2) * self.Om(xx) * self.f_intp(xx))
        beta[18] = lambda xx: self.H_c(xx) * ( (3 / 2) * self.Om(xx) * self.f_intp(xx) - self.f_intp(xx)**2 * (3 - 2 * self.b_e(xx) + 4 * self.Q(xx) + ((4 * (1 - self.Q(xx))) / (self.comoving_dist(xx) * self.H_c(xx))) + (3 * dH_dt(xx)/ self.H_c(xx)**2)) )
        beta[19] = lambda xx: self.H_c(xx) * (self.f_intp(xx)* (self.b_e(xx) - 2 * self.Q(xx) - ((2 * (1 - self.Q(xx))) / (self.comoving_dist(xx) * self.H_c(xx))) - (dH_dt(xx)/ self.H_c(xx)**2)))

        
        #get betad - derivatives wrt to ln(d)  
        betad = np.empty(20,dtype=object)
        betad[14:20] = np.array(self.lnd_derivatives(beta[14:20]),dtype=object)#14-19
        
        tracer.beta = beta 
        tracer.betad = betad
        all_betas = np.concatenate((np.array([tracer.gr1,tracer.gr2,tracer.grd1]),beta[6:],betad[14:20]))
        return [all_betas[i](zz) for i in range(len(all_betas))]
    



import numpy as np

def RR_1(params,derivs,r,s,mu,phi):
    k1,k2,k3,theta,Pk1,Pk2,Pk3,Pkd1,Pkd2,Pkd3,Pkdd1,Pkdd2,Pkdd3,d,K,C,f,D1,b1,b2,g2 = params
    fd,Dd,gd2,bd2,bd1,fdd,Ddd,gdd2,bdd2,bdd1 = derivs

    mu2 = mu*np.cos(theta)+(1-mu)**2 *np.sin(theta)*np.cos(phi)
    st = np.sin(theta)
    ct= np.cos(theta)

    perm12 = 1
    perm13 = 1
    perm23 = 1
    return (perm12+perm13+perm23)